//! Lagging iterators / (todo) streams.
use super::*;
use std::time::Duration;

/// A lagged iterator
#[derive(Debug, Clone)]
pub struct Lag<I: ?Sized, T>(Duration, I)
where I: Iterator<Item=T>;

impl<I: ?Sized, T>  Lag<I, T>
where I: Iterator<Item=T>
{
    /// Set the lag duration
    #[inline] pub fn set_duration(&mut self, dur: Duration)
    {
	self.0 = dur;
    }
    /// Get the lag duration
    #[inline] pub fn duration(&self) -> Duration
    {
	self.0
    }
}
impl<I, T>  Lag<I, T>
where I: Iterator<Item=T>
{
    /// Consume into the inner iterator
    #[inline] pub fn into_inner(self) -> I
    {
	self.1
    }
}

pub trait LagIterExt: Iterator
{
    fn lag(self, dur: Duration) -> Lag<Self, Self::Item>;
}

impl<I> LagIterExt for I
where I: Iterator
{
    #[inline] fn lag(self, dur: Duration) -> Lag<Self, Self::Item>
    {
	Lag(dur, self)
    }
}

impl<I: ?Sized, T> Iterator for Lag<I, T>
where I: Iterator<Item=T>
{
    type Item = T;
    fn next(&mut self) -> Option<Self::Item>
    {
	std::thread::sleep(self.0);
	self.1.next()
    }
    fn size_hint(&self) -> (usize, Option<usize>) {
	self.1.size_hint()
    }
}

impl<I: ?Sized, T> FusedIterator for Lag<I, T>
where I: Iterator<Item=T> + FusedIterator{}
impl<I: ?Sized, T> ExactSizeIterator for Lag<I, T>
where I: Iterator<Item=T> + ExactSizeIterator{}
impl<I: ?Sized, T> DoubleEndedIterator for Lag<I, T>
where I: Iterator<Item=T> + DoubleEndedIterator
{
    fn next_back(&mut self) -> Option<Self::Item>
    {
	std::thread::sleep(self.0);
	self.1.next()
    }
}
